#include <sys/time.h>
#include <event.h>
#include <sys/socket.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>


#include <konoha.h>

#if (KONOHA_BUILDID <= 771)
#define knh_IntData_t knh_IntConstData_t
#define knh_loadIntConst knh_loadIntConstData
#define knh_StreamDSPI_t knh_iodrv_t
#endif


#ifdef __cplusplus
extern "C" {
#endif
static knh_IntData_t IntConstData[] = {
	{"Event.EV_TIMEOUT",EV_TIMEOUT},
	{"Event.EV_READ",EV_READ},
	{"Event.EV_WRITE",EV_WRITE},
	{"Event.EV_SIGNAL",EV_SIGNAL},
	{"Event.EV_PERSIST",EV_PERSIST},
	{NULL} // end of const data
};


typedef struct knh_event_t {
	struct event e;
	knh_Closure_t *func;
	Object* obj;
} knh_event_t;

static void knh_ev_gfree(Ctx* ctx, knh_Glue_t *g)
{
	knh_event_t *ev = (knh_event_t *)g->ptr;
	KNH_FINALv(ctx, ev->func);
	KNH_FINALv(ctx, ev->obj);
	free(g->ptr);
}


static void knh_event(int fd, short flag, void *arg)
{
	knh_event_t *ev = (knh_event_t *)arg;
	Ctx *ctx = knh_getCurrentContext();
	knh_sfp_t *lsfp = KNH_LOCAL(ctx);

	knh_Closure_t *cc = ev->func;
	Object *arg1 = ev->obj;
	if(IS_Socket(arg1) && (flag & EV_READ)) {
		int sd = accept(fd, NULL, NULL);
		knh_Socket_t *o = new_Socket(ctx, sd);
		knh_StreamDSPI_t *drv = knh_getSocketDriver();
		DP(o)->sd = sd;
		DP(o)->port = DP(o)->port;
		KNH_SETv(ctx, DP(o)->urn, DP((knh_Socket_t*)arg1)->urn);
		knh_InputStream_t *in = new_InputStream__io(ctx, DP(o)->urn, sd, drv);
		knh_OutputStream_t *out = new_OutputStream__io(ctx, DP(o)->urn, sd, drv);
		KNH_SETv(ctx, DP(o)->in,  in);
		KNH_SETv(ctx, DP(o)->out, out);
		knh_putsfp(ctx,lsfp,2,UP(o));
	} else {
		knh_putsfp(ctx,lsfp,2,arg1);
	}
	knh_Closure_invokesfp(ctx,cc,lsfp,1);
}

// This! Event.new();
METHOD Event_new(Ctx *ctx, knh_sfp_t *sfp)
{
	static int knh_libevent_is_init = 0;
	if(knh_libevent_is_init == 0) {
		event_init();
		knh_libevent_is_init = 1;
	}
	knh_event_t *ev = (knh_event_t *) malloc(sizeof(*ev));
	KNH_INITv(ev->func, KNH_NULL);
	KNH_INITv(ev->obj, KNH_NULL);

	knh_Glue_init(ctx, sfp[0].glue, (void *)ev, knh_ev_gfree);
	KNH_RETURN(ctx, sfp, sfp[0].o);
}

// void Event.newEvent(int flag, Object object, Closure callback);
METHOD Event_newEvent(Ctx *ctx, knh_sfp_t *sfp)
{
	knh_event_t* ev = (knh_event_t *) p_cptr(sfp[0]);
	int flag = p_int(sfp[1]);
	Object *o = sfp[2].o;
	int fd = -1;
	if(IS_OutputStream(o) || IS_InputStream(o)) {
		fd = DP((knh_InputStream_t*)o)->fd;
	} else if(IS_Socket(o)) {
		fd = DP((knh_Socket_t*)o)->sd;
	}
	knh_Closure_t *cc = sfp[3].cc;
	KNH_SETv(ctx, ev->func, cc);
	KNH_SETv(ctx, ev->obj , o);

	event_set(&ev->e, fd, flag, knh_event, ev);
	KNH_SETv(ctx, ev->func, cc);
	KNH_RETURN_void(ctx, sfp);
}

// void Event.add:NoTimeout();
METHOD Event_add2(Ctx *ctx, knh_sfp_t *sfp)
{
	knh_event_t* ev = (knh_event_t *) p_cptr(sfp[0]);
	event_add(&ev->e, NULL);
	KNH_RETURN_void(ctx, sfp);
}

// void Event.add(float timeout);
METHOD Event_add(Ctx *ctx, knh_sfp_t *sfp)
{
	knh_event_t* ev = (knh_event_t *) p_cptr(sfp[0]);
	float timeout = p_float(sfp[1]);
	struct timeval tval;
	tval.tv_sec  = timeout;
	tval.tv_usec = 0;
	event_add(&ev->e, &tval);
	KNH_RETURN_void(ctx, sfp);
}

// void Event.dispatch();
METHOD Event_dispatch(Ctx *ctx, knh_sfp_t *sfp)
{
	event_dispatch();
	KNH_RETURN_void(ctx, sfp);
}

KNH_EXPORTS(int) init(Ctx *ctx)
{
	KNH_NOTICE(ctx, "loading event..");
	knh_loadIntConstData(ctx, IntConstData);
	return 1;
}

#ifdef __cplusplus
}
#endif

